/*-
 * Copyright (c) 2008, Atmel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * clock.c - Atmel at91 TC clock driver
 */

#include <kernel.h>
#include <timer.h>
#include <irq.h>
#include <cpufunc.h>
#include <sys/ipl.h>

#include "platform.h"

/* Peripheral Clock Enable Register */
#define PMC_PCER (*(volatile uint32_t *)(0xfffffc10)) /* Peripheral Clock Enable Register */

/* Peripheral id TC0 */
#define ID_TC  17

/* Timer Counter registers */
#define TC_CCR (*(volatile uint32_t *)(TC_BASE + 0x0000)) /* Channel Control Register */
#define TC_CMR (*(volatile uint32_t *)(TC_BASE + 0x0004)) /* Channel Mode Register */
#define TC_RC  (*(volatile uint32_t *)(TC_BASE + 0x001c)) /* Register C */
#define TC_SR  (*(volatile uint32_t *)(TC_BASE + 0x0020)) /* Status Register */
#define TC_IER (*(volatile uint32_t *)(TC_BASE + 0x0024)) /* Interrupt Enable Register */
#define TC_IDR (*(volatile uint32_t *)(TC_BASE + 0x0028)) /* Interrupt Disable Register */

#define CCR_CLKEN              (0x1 <<  0) /* Counter Clock Enable Command */
#define CCR_CLKDIS             (0x1 <<  1) /* Counter Clock Disable Command */
#define CCR_SWTRG              (0x1 <<  2) /* Software Trigger Command */
#define CMR_WAVE               (0x1 << 15)
#define CMR_WAVESEL_UP_AUTO    (0x2 << 13) /* UP mode with automatic trigger on RC Compare */
#define SR_CPCS                (0x1 <<  4) /* RC Compare */

/*
 * Finds the best MCK divisor given the timer frequency and MCK
 */
unsigned char
find_divisor(unsigned int freq,
             unsigned int *div,
             unsigned int *tcclks)
{
    const unsigned int divisors[5] = {2, 8, 32, 128, CONFIG_MCK / 32768};
    unsigned int index = 0;

    /* Satisfy lower bound */
    while (freq < ((CONFIG_MCK / divisors[index]) / 65536)) {
        index++;
        /* If no divisor can be found, return 0 */
        if (index == 5)
            return 0;
    }

    /* Try to maximise DIV while satisfying upper bound */
    while (index < 4) {
        if (freq > (CONFIG_MCK / divisors[index + 1]))
            break;
        index++;
    }

    /* Store results */
    if (div)
        *div = divisors[index];
    if (tcclks)
        *tcclks = index;
    return 1;
}

/*
 * Clock interrupt service routine.
 * No H/W reprogram is required.
 */
static int
clock_isr(void *arg)
{
    volatile uint32_t piv;

    /* Acknowledging the interrupt */
    piv = TC_SR;
    splhigh();
    timer_handler();
    spl0();
    return INT_DONE;
}

/*
 * Initialize clock H/W chip.
 * Setup clock tick rate and install clock ISR.
 */
void
clock_init(void)
{
    irq_t clock_irq;
    unsigned int div, tcclks, sr;

    /* Enable TC */
    PMC_PCER = 1 << ID_TC;

    find_divisor(CONFIG_HZ, &div, &tcclks);
    /* Disable TC clock */
    TC_CCR = CCR_CLKDIS;
    /* Disable interrupts */
    TC_IDR = 0xffffffff;
    /* Clear status register */
    sr = TC_SR;
    /* Set mode */
    TC_CMR = tcclks | CMR_WAVE | CMR_WAVESEL_UP_AUTO;
    TC_RC = CONFIG_MCK / (CONFIG_HZ * div);

    clock_irq = irq_attach(ID_TC, IPL_CLOCK, 1, &clock_isr, IST_NONE, NULL);

    /* Enable timer interrupt */
    TC_IER = SR_CPCS;
    /* Start timer */
    TC_CCR = CCR_CLKEN | CCR_SWTRG;

    DPRINTF(("Clock rate: %d ticks/sec\n", CONFIG_HZ));
}


